--[[
CastYeller2 - Announce important spell events.

Copyright (C) 2009-2011 Adirelle

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
--]]

if not CastYeller2 then return end
local CastYeller2 = CastYeller2 

--------------------------------------------------------------------------------
-- Locals
--------------------------------------------------------------------------------

local baseModulePrototype, sourceModulePrototype, sinkModulePrototype
local combatCallbacks

--------------------------------------------------------------------------------
-- Base module prototype
--------------------------------------------------------------------------------

baseModulePrototype = { 	
	core = CastYeller2,
	RegisterSignal = CastYeller2.RegisterSignal, 
	UnregisterSignal = CastYeller2.UnregisterSignal,
	UnregisterAllSignals = CastYeller2.UnregisterAllSignals,
	SendSignal = CastYeller2.SendSignal,
	Debug = CastYeller2.Debug,
}

function baseModulePrototype:OnInitialize()
	CastYeller2:RegisterModule(self)
end

function baseModulePrototype:OnEnable()
end

function baseModulePrototype:OnDisable()
end

CastYeller2:SetDefaultModuleState(true)
CastYeller2:SetDefaultModulePrototype(baseModulePrototype)

--------------------------------------------------------------------------------
-- Source module prototype
--------------------------------------------------------------------------------

sourceModulePrototype = setmetatable({ moduleType = 'source' }, { __index = baseModulePrototype })
CastYeller2.sourceModulePrototype = sourceModulePrototype

function sourceModulePrototype:OnDisable()
	baseModulePrototype.OnDisable(self)
	self:UnregisterAllCombatListeners()
end

function sourceModulePrototype:FindAnnounceFor(trigger, spell)
	local inCombat = CastYeller2.testMode or InCombatLockdown()
	for name, announce in self.core:IterateActiveAnnounces() do
		if announce.sources[self.moduleName] and announce.triggers[trigger] and announce.names[spell] and (inCombat or not announce.inCombat) then
			return announce
		end
	end
end

function sourceModulePrototype:IsNameWatched(name) 
	return self.core:IsNameWatched(name, self.moduleName) 
end
	
function sourceModulePrototype:IterateWatchedNames()
	return self.core:IterateWatchedNames(self.moduleName) 
end

do
	local icons = {}
	for i = 1,8 do icons[_G['COMBATLOG_OBJECT_RAIDTARGET'..i]] = i end
	function sourceModulePrototype:GetRaidIcon(flags)
		return icons[bit.band(tonumber(flags) or 0, COMBATLOG_OBJECT_RAIDTARGET_MASK)]
	end
end

function sourceModulePrototype:TriggerAnnounce(trigger, target, spell, ...)	
	if not self:IsNameWatched(spell) then return end
	local announce = self:FindAnnounceFor(trigger, spell)
	if announce and not self:VetoAnnounce(announce, trigger, target, spell, ...) then
		local info = self:BuildInfo(announce, trigger, target, spell, ...)	
		if info.raidIcon then
			info.raidIconTexture = '{rt'.. info.raidIcon..'}'
		end
		self:SendSignal('OnAnnounceSent', announce, info)
	end
end

function sourceModulePrototype:VetoAnnounce()
	return false
end

do 
	local info = {}
	function sourceModulePrototype:BuildInfo(announce, trigger, target, spell, duration, spellId, targetGUID, targetFlags, ...)
		wipe(info)
		info.source = self.moduleName
		info.trigger = trigger
		if target and target ~= UNKNOWN then
			info.target = target
			info.targetGUID = targetGUID
			info.targetFlags = targetFlags
		end
		if duration then
			local fmt, t = SecondsToTimeAbbrev(duration)
			info.duration = fmt:format(t)
		end
		info.spell = spell
		info.spellId = spellId
		info.spellLink = GetSpellLink(spellId or spell or "")
		for i = 1, select('#', ...), 2 do
			info[select(i, ...)] = select(i+1, ...)
		end
		if info.raidIcon == nil then			
			if targetFlags then
				info.raidIcon = self:GetRaidIcon(targetFlags)		
			elseif target and (UnitInParty(target) or UnitInRaid(target)) then
				info.raidIcon = GetRaidTargetIndex(target)					
			end
		end
		return info
	end
end

--------------------------------------------------------------------------------
-- Sink module prototype
--------------------------------------------------------------------------------

sinkModulePrototype = setmetatable({ moduleType = 'sink' }, { __index = baseModulePrototype })
CastYeller2.sinkModulePrototype = sinkModulePrototype

function sinkModulePrototype:OnEnable()
	baseModulePrototype.OnEnable(self)
	self:RegisterSignal("OnAnnounceSent")
end

function sinkModulePrototype:OnDisable()
	baseModulePrototype.OnDisable(self)
	self:UnregisterSignal("OnAnnounceSent")
end

function sinkModulePrototype:OnAnnounceSent(event, announce, info)
	if announce.sinks[self.moduleName] then
		self:Sink(announce, info)
	end
end

function sinkModulePrototype:GetMessageFormat(announce, info, prefix)
	local trigger, target = info.trigger, info.target
	local custom
	if announce.custom[trigger] and announce.custom[trigger].enabled then
		local announces = announce.custom[trigger].announces
		if type(announces) == "table" then
			local t = (prefix and announces[prefix]) or announces['*']
			if t and #t > 0 then
				return t[math.random(1, #t)]
			end
		end
	end
	return self.core.db.profile.templates[info.source][trigger]
end

function sinkModulePrototype:FormatMessage(announce, info, prefix, noBrackets, noLinks)
	local template = self:GetMessageFormat(announce, info, prefix)
	if not template then 
		-- Should not happen
		print('CastYeller2 error: could not find a template for trigger ', info.source..'.'..info.trigger)
		return ""
	end
	local oldSpell = info.spell
	if not noLinks and info.spellLink and self.core.db.profile.useSpellLinks then
		info.spell = info.spellLink
	end
	local message = self.core:FormatMessage(template, info)
	info.spell = oldSpell
	if not noBrackets and self.core.db.profile.squareBrackets then
		message = '<< '..message..' >>'
	end
	return message
end

--------------------------------------------------------------------------------
-- Combat listener handling
--------------------------------------------------------------------------------

local registeredEvents = {}
combatCallbacks = LibStub("CallbackHandler-1.0"):New(
	sourceModulePrototype, 
  "RegisterCombatListener",
  "UnregisterCombatListener",
  "UnregisterAllCombatListeners"
)
LibStub('AceEvent-3.0'):Embed(combatCallbacks)

function combatCallbacks:OnUsed(target, event)
	if not next(registeredEvents) then		
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
	registeredEvents[event] = true
end

function combatCallbacks:OnUnused(target, event)
	registeredEvents[event] = nil
	if not next(registeredEvents) then
		self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
end

function combatCallbacks:COMBAT_LOG_EVENT_UNFILTERED(_, time, event, hideCaster, ...)
	if registeredEvents[event] then
		self:Fire(event, time, ...)
	end
end

-- 4.2 compat layer
if select(4, GetBuildInfo()) == 40200 then
	local Base_COMBAT_LOG_EVENT_UNFILTERED = combatCallbacks.COMBAT_LOG_EVENT_UNFILTERED
	function combatCallbacks.COMBAT_LOG_EVENT_UNFILTERED(self, _, time, event, hideCaster, sourceGUID, sourceName, sourceFlags, _, destGUID, destName, destFlags, _, ...)
		return Base_COMBAT_LOG_EVENT_UNFILTERED(self, _, time, event, hideCaster, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
	end
end

